@@ -33,7 +33,7 @@
 #ifdef HAVE_CAP_GET_PROC
 
 static int
-bin_cap(char *nam, char **argv, Options ops, int func)
+bin_cap(char *nam, char **argv, UNUSED(Options ops), UNUSED(int func))
 {
     int ret = 0;
     cap_t caps;
@@ -64,7 +64,7 @@ bin_cap(char *nam, char **argv, Options ops, int func)
 }
 
 static int
-bin_getcap(char *nam, char **argv, Options ops, int func)
+bin_getcap(char *nam, char **argv, UNUSED(Options ops), UNUSED(int func))
 {
     int ret = 0;
 
@@ -85,7 +85,7 @@ bin_getcap(char *nam, char **argv, Options ops, int func)
 }
 
 static int
-bin_setcap(char *nam, char **argv, Options ops, int func)
+bin_setcap(char *nam, char **argv, UNUSED(Options ops), UNUSED(int func))
 {
     cap_t caps;
     int ret = 0;
