@@ -337,26 +337,22 @@ static void
 add_match_part(Cmatcher m, char *l, char *w, int wl,
 	       char *o, int ol, char *s, int sl, int osl, int sfx)
 {
-    Cline p, lp, lprem;
+    Cline p, lp;
 
     /* If the anchors are equal, we keep only one. */
 
     if (l && !strncmp(l, w, wl))
 	l = NULL;
 
-    /*
-     * Split the new part into parts and turn the last one into a
-     * `suffix' if we have a left anchor---don't do this if the last one
-     * came from a right anchor before the end of the part we're
-     * splitting.
-     */
+    /* Split the new part into parts and turn the last one into a
+     * `suffix' if we have a left anchor. */
 
-    p = bld_parts(s, sl, osl, &lp, &lprem);
+    p = bld_parts(s, sl, osl, &lp);
 
-    if (lprem && m && (m->flags & CMF_LEFT)) {
-	lprem->flags |= CLF_SUF;
-	lprem->suffix = lprem->prefix;
-	lprem->prefix = NULL;
+    if (m && (m->flags & CMF_LEFT)) {
+	lp->flags |= CLF_SUF;
+	lp->suffix = lp->prefix;
+	lp->prefix = NULL;
     }
     /* cline lists for suffixes are sorted from back to front, so we have
      * to revert the list we got. */
@@ -422,7 +418,7 @@ add_match_sub(Cmatcher m, char *l, int ll, char *w, int wl)
     if (wl || ll) {
 	Cline p, lp;
 
-	if ((p = n = bld_parts(w, wl, ll, &lp, NULL)) && n != lp) {
+	if ((p = n = bld_parts(w, wl, ll, &lp)) && n != lp) {
 	    for (; p->next != lp; p = p->next);
 
 	    if (matchsubs) {
@@ -973,8 +969,7 @@ match_parts(char *l, char *w, int n, int part)
 /* Check if the word w is matched by the strings in pfx and sfx (the prefix
  * and the suffix from the line) or the pattern cp. In clp a cline list for
  * w is returned.
- * qu is non-zero if the words has to be quoted before processed any
- * further: the value 2 indicates file quoting.
+ * qu is non-zero if the words has to be quoted before processed any further.
  * bpl and bsl are used to report the positions where the brace-strings in
  * the prefix and the suffix have to be re-inserted if this match is inserted
  * in the line.
@@ -991,30 +986,9 @@ comp_match(char *pfx, char *sfx, char *w, Patprog cp, Cline *clp, int qu,
     if (cp) {
 	/* We have a globcomplete-like pattern, just use that. */
 	int wl;
-	char *teststr;
 
 	r = w;
-	if (!qu) {
-	    /*
-	     * If we're not quoting the strings, that means they're
-	     * already quoted (?) and so when we test them against
-	     * a pattern we have to remove the quotes else we will
-	     * end up trying to match against the quote characters.
-	     *
-	     * Almost certainly this fails in some complicated cases
-	     * but it should catch the basic ones.
-	     */
-	    teststr = dupstring(r);
-	    tokenize(teststr);
-	    if (parse_subst_string(teststr))
-		teststr = r;
-	    else {
-		remnulargs(teststr);
-		untokenize(teststr);
-	    }
-	} else
-	    teststr = r;
-	if (!pattry(cp, teststr))
+	if (!pattry(cp, r))
 	    return NULL;
     
 	r = (qu == 2 ? tildequote(r, 0) : multiquote(r, !qu));
@@ -1023,7 +997,7 @@ comp_match(char *pfx, char *sfx, char *w, Patprog cp, Cline *clp, int qu,
 	 * cline list for these matches, too. */
 	w = dupstring(w);
 	wl = strlen(w);
-	*clp = bld_parts(w, wl, wl, NULL, NULL);
+	*clp = bld_parts(w, wl, wl, NULL);
 	*exact = 0;
     } else {
 	Cline pli, plil;
@@ -1081,7 +1055,7 @@ comp_match(char *pfx, char *sfx, char *w, Patprog cp, Cline *clp, int qu,
 	    add_match_str(NULL, NULL, wpfx, wpl, 1);
 
 	    mli = bld_parts(w + rpl, wl - rpl - rsl,
-			    (mpl - rpl) + (msl - rsl), &mlil, NULL);
+			    (mpl - rpl) + (msl - rsl), &mlil);
 	    mlil->flags |= CLF_MID;
 	    mlil->slen = msl - rsl;
 	    mlil->next = revert_cline(matchparts);
@@ -1161,19 +1135,11 @@ pattern_match(Cpattern p, char *s, unsigned char *in, unsigned char *out)
 /* This splits the given string into a list of cline structs, separated
  * at those places where one of the anchors of an `*' pattern was found.
  * plen gives the number of characters on the line that matched this
- * string.
- *
- * In *lp, if lp is not NULL, we return a pointer to the last cline struct we
- * build.
- *
- * In *lprem, if lprem is not NULL, we return a pointer to the last
- * cline struct we build if it came from the remainder of the
- * line rather than from a right anchor match, else NULL.
- */
+ * string. In lp we return a pointer to the last cline struct we build. */
 
 /**/
 Cline
-bld_parts(char *str, int len, int plen, Cline *lp, Cline *lprem)
+bld_parts(char *str, int len, int plen, Cline *lp)
 {
     Cline ret = NULL, *q = &ret, n = NULL;
     Cmlist ms;
@@ -1232,17 +1198,9 @@ bld_parts(char *str, int len, int plen, Cline *lp, Cline *lprem)
 	if (llen > olen)
 	    llen = olen;
 	n->prefix = get_cline(NULL, llen, p, olen, NULL, 0, 0);
-	if (lprem)
-	    *lprem = n;
-    }
-    else if (!ret) {
-        *q = n =
-	    get_cline(NULL, 0, NULL, 0, NULL, 0, (plen <= 0 ? CLF_NEW : 0));
-	if (lprem)
-	    *lprem = n;
-    } else if (lprem) {
-	*lprem = NULL;
     }
+    else if (!ret)
+        *q = n = get_cline(NULL, 0, NULL, 0, NULL, 0, (plen <= 0 ? CLF_NEW : 0));
 
     n->next = NULL;
 
@@ -1626,6 +1584,11 @@ sub_match(Cmdata md, char *str, int len, int sfx)
 {
     int ret = 0, l, ind, add;
     char *p, *q;
+#ifdef MULTIBYTE_SUPPORT
+    int fulllen = len;
+    char *fullstr = str;
+    mbstate_t mbs;
+#endif
 
     if (sfx) {
 	str += len;
@@ -1656,6 +1619,84 @@ sub_match(Cmdata md, char *str, int len, int sfx)
 		   || (l < md->len && q[-1] == Meta)))
 		l--;
 	}
+#ifdef MULTIBYTE_SUPPORT
+	/*
+	 * Make sure we don't end in the middle of a multibyte character.
+	 * Don't need to do this if the match ended at the start
+	 * of the original string.
+	 *
+	 * Let q be the match point we've found.
+	 */
+	q = sfx ? str - l : str + l;
+	if (q != fullstr) {
+	    memset(&mbs, 0, sizeof mbs);
+	    /*
+	     * Otherwise read characters from the start of the original
+	     * string until we reach or pass the match point.  This
+	     * is rather inefficient, but in general only reading
+	     * the full string can keep track of where we are in
+	     * a character.  With a prefix we could be more efficient,
+	     * but it's difficult with a suffix where the match point
+	     * moves backwards.
+	     */
+	    for (p = fullstr; p < fullstr + fulllen; ) {
+		wchar_t wc;
+		/*
+		 * ret must, in fact, be set by the current logic,
+		 * but gcc doesn't realise (at least some versions don't).
+		 */
+		size_t cnt = MB_INVALID;
+		int diff;
+		char *p2;
+
+		/*
+		 * Because the string is metafied, we need to
+		 * assembled wide characters a byte at a time.
+		 */
+		for (p2 = p; p2 < fullstr + fulllen; p2++) {
+		    char curchar = (*p2 == Meta) ? (*++p2 ^ 32) : *p2;
+		    cnt = mbrtowc(&wc, &curchar, 1, &mbs);
+		    /* Continue while character is incomplete. */
+		    if (cnt != MB_INCOMPLETE)
+			break;
+		}
+		if (cnt == MB_INVALID || cnt == MB_INCOMPLETE) {
+		    /* not a valid character, give up test */
+		    break;
+		}
+		/* increment p2 for last byte read */
+		diff = ++p2 - q;
+		if (diff == 0) {
+		    /*
+		     * Prefix or suffix matches at end of multbyte character,
+		     * so OK.
+		     */
+		    break;
+		} else if (diff > 0) {
+		    /*
+		     * The prefix or suffix finishes in the middle
+		     * of a character.  Shorten it until it doesn't.
+		     */
+		    if (sfx) {
+			/*
+			 * We need to remove the trailing part of
+			 * the character from the suffix.
+			 */
+			l -= diff;
+		    } else {
+			/*
+			 * We need to remove the initial part of
+			 * the character from the prefix.
+			 */
+			l -= (q - p);
+		    }
+		    break;
+		}
+		/* Advance over full character */
+		p = p2;
+	    }
+	}
+#endif
 	if (l) {
 	    /* There was a common prefix, use it. */
 	    md->len -= l; len -= l;
