@@ -77,12 +77,12 @@ mod_export int tclen[TC_COUNT];
 /**/
 int tclines, tccolumns;
 /**/
-mod_export int hasam;
+mod_export int hasam, hasxn;
 
 /* Pointer to read-key function from zle */
 
 /**/
-mod_export int (*getkeyptr) _((int));
+mod_export int (*getkeyptr) _((int, int *));
 
 /* SIGCHLD mask */
 
@@ -131,15 +131,6 @@ loop(int toplevel, int justonce)
 		(tok == LEXERR && (!isset(SHINSTDIN) || !toplevel)) ||
 		justonce)
 		break;
-	    if (exit_pending) {
-		/*
-		 * Something down there (a ZLE function?) decided
-		 * to exit when there was stuff to clear up.
-		 * Handle that now.
-		 */
-		stopmsg = 1;
-		zexit(exit_pending >> 1, 0);
-	    }
 	    if (tok == LEXERR && !lastval)
 		lastval = 1;
 	    continue;
@@ -527,7 +518,7 @@ static char *tccapnams[TC_COUNT] = {
     "cl", "le", "LE", "nd", "RI", "up", "UP", "do",
     "DO", "dc", "DC", "ic", "IC", "cd", "ce", "al", "dl", "ta",
     "md", "so", "us", "me", "se", "ue", "ch",
-    "ku", "kd", "kl", "kr"
+    "ku", "kd", "kl", "kr", "sc", "rc", "bc"
 };
 
 /* Initialise termcap */
@@ -582,6 +573,7 @@ init_term(void)
 
 	/* check whether terminal has automargin (wraparound) capability */
 	hasam = tgetflag("am");
+	hasxn = tgetflag("xn"); /* also check for newline wraparound glitch */
 
 	tclines = tgetnum("li");
 	tccolumns = tgetnum("co");
@@ -596,10 +588,22 @@ init_term(void)
 	    termflags |= TERM_NOUP;
 	}
 
-	/* if there's no termcap entry for cursor left, use \b. */
+	/* most termcaps don't define "bc" because they use \b. */
+	if (!tccan(TCBACKSPACE)) {
+	    tcstr[TCBACKSPACE] = ztrdup("\b");
+	    tclen[TCBACKSPACE] = 1;
+	}
+
+	/* if there's no termcap entry for cursor left, use backspace. */
 	if (!tccan(TCLEFT)) {
-	    tcstr[TCLEFT] = ztrdup("\b");
-	    tclen[TCLEFT] = 1;
+	    tcstr[TCLEFT] = tcstr[TCBACKSPACE];
+	    tclen[TCLEFT] = tclen[TCBACKSPACE];
+	}
+
+	if (tccan(TCSAVECURSOR) && !tccan(TCRESTRCURSOR)) {
+	    tclen[TCSAVECURSOR] = 0;
+	    zsfree(tcstr[TCSAVECURSOR]);
+	    tcstr[TCSAVECURSOR] = NULL;
 	}
 
 	/* if the termcap entry for down is \n, don't use it. */
@@ -624,7 +628,7 @@ init_term(void)
 void
 setupvals(void)
 {
-#ifdef HAVE_GETPWUID
+#ifdef USE_GETPWUID
     struct passwd *pswd;
 #endif
     struct timezone dummy_tz;
@@ -791,22 +795,32 @@ setupvals(void)
      * recheck the info for `USERNAME'     */
     cached_uid = getuid();
 
-    /* Get password entry and set info for `HOME' and `USERNAME' */
-#ifdef HAVE_GETPWUID
+    /* Get password entry and set info for `USERNAME' */
+#ifdef USE_GETPWUID
     if ((pswd = getpwuid(cached_uid))) {
-	home = metafy(pswd->pw_dir, -1, META_DUP);
+	if (emulation == EMULATE_ZSH)
+	    home = metafy(pswd->pw_dir, -1, META_DUP);
 	cached_username = ztrdup(pswd->pw_name);
-    } else
-#endif /* HAVE_GETPWUID */
-	   {
-	home = ztrdup("/");
+    }
+    else
+#endif /* USE_GETPWUID */
+    {
+	if (emulation == EMULATE_ZSH)
+	    home = ztrdup("/");
 	cached_username = ztrdup("");
     }
 
-    /* Try a cheap test to see if we can *
-     * initialize `PWD' from `HOME'      */
-    if (ispwd(home))
-	pwd = ztrdup(home);
+    /*
+     * Try a cheap test to see if we can initialize `PWD' from `HOME'.
+     * In non-native emulations HOME must come from the environment;
+     * we're not allowed to set it locally.
+     */
+    if (emulation == EMULATE_ZSH)
+	ptr = home;
+    else
+	ptr = getenv("HOME");
+    if (ptr && ispwd(ptr))
+	pwd = ztrdup(ptr);
     else if ((ptr = zgetenv("PWD")) && (strlen(ptr) < PATH_MAX) &&
 	     (ptr = metafy(ptr, -1, META_STATIC), ispwd(ptr)))
 	pwd = ztrdup(ptr);
@@ -944,8 +958,20 @@ run_init_scripts(void)
 #ifdef GLOBAL_ZSHENV
 	source(GLOBAL_ZSHENV);
 #endif
+
 	if (isset(RCS) && unset(PRIVILEGED))
+	{
+	    /*
+	     * Always attempt to load the newuser module to perform
+	     * checks for new zsh users.  Don't care if we can't load it.
+	     */
+	    if (load_module_silence("zsh/newuser", 1)) {
+		/* Unload it immediately. */
+		unload_named_module("zsh/newuser", "zsh", 1);
+	    }
+
 	    sourcehome(".zshenv");
+	}
 	if (islogin) {
 #ifdef GLOBAL_ZPROFILE
 	    if (isset(RCS) && isset(GLOBALRCS))
@@ -1089,8 +1115,11 @@ sourcehome(char *s)
 
     queue_signals();
     if (emulation == EMULATE_SH || emulation == EMULATE_KSH ||
-	!(h = getsparam("ZDOTDIR")))
+	!(h = getsparam("ZDOTDIR"))) {
 	h = home;
+	if (!h)
+	    return;
+    }
 
     {
 	/* Let source() complain if path is too long */
@@ -1137,9 +1166,11 @@ mod_export ZleVoidFn trashzleptr = noop_function;
 /**/
 mod_export ZleVoidFn zle_resetpromptptr = noop_function;
 /**/
-mod_export ZleVoidFn refreshptr = noop_function;
+mod_export ZleVoidFn zrefreshptr = noop_function;
 /**/
-mod_export ZleVoidIntFn spaceinlineptr = noop_function_int;
+mod_export ZleVoidIntFn zleaddtolineptr = noop_function_int;
+/**/
+mod_export ZleGetLineFn zlegetlineptr = NULL;
 /**/
 mod_export ZleReadFn zlereadptr = autoload_zleread;
 /**/
@@ -1149,8 +1180,9 @@ mod_export ZleVoidIntFn zlesetkeymapptr = noop_function_int;
 
 mod_export ZleVoidFn trashzleptr = noop_function;
 mod_export ZleVoidFn zle_resetpromptptr = noop_function;
-mod_export ZleVoidFn refreshptr = noop_function;
-mod_export ZleVoidIntFn spaceinlineptr = noop_function_int;
+mod_export ZleVoidFn zrefreshptr = noop_function;
+mod_export ZleVoidIntFn zleaddtolineptr = noop_function_int;
+mod_export ZleGetLineFn zlegetlineptr = NULL;
 # ifdef UNLINKED_XMOD_zshQszle
 mod_export ZleReadFn zlereadptr = autoload_zleread;
 mod_export ZleVoidIntFn zlesetkeymapptr = autoload_zlesetkeymap;
@@ -1162,17 +1194,17 @@ mod_export ZleVoidIntFn zlesetkeymapptr = noop_function_int;
 #endif /* !LINKED_XMOD_zshQszle */
 
 /**/
-unsigned char *
+char *
 autoload_zleread(char **lp, char **rp, int ha, int con)
 {
     zlereadptr = fallback_zleread;
     if (load_module("zsh/zle"))
 	load_module("zsh/compctl");
-    return zleread(lp, rp, ha, con);
+    return zlereadptr(lp, rp, ha, con);
 }
 
 /**/
-mod_export unsigned char *
+mod_export char *
 fallback_zleread(char **lp, UNUSED(char **rp), UNUSED(int ha), UNUSED(int con))
 {
     char *pptbuf;
@@ -1182,9 +1214,12 @@ fallback_zleread(char **lp, UNUSED(char **rp), UNUSED(int ha), UNUSED(int con))
     write(2, (WRITE_ARG_2_T)pptbuf, pptlen);
     free(pptbuf);
 
-    return (unsigned char *)shingetline();
+    return shingetline();
 }
 
+/**/
+#ifdef UNLINKED_XMOD_zshQszle
+
 /**/
 static void
 autoload_zlesetkeymap(int mode)
@@ -1194,6 +1229,8 @@ autoload_zlesetkeymap(int mode)
     (*zlesetkeymapptr)(mode);
 }
 
+/**/
+#endif
 
 /* compctl entry point pointers.  Similar to the ZLE ones. */
 
@@ -1285,11 +1322,9 @@ zsh_main(UNUSED(int argc), char **argv)
 	 */
 	maybeshrinkjobtab();
 
-	do {
-	    /* Reset return from top level which gets us back here */
-	    retflag = 0;
+	do
 	    loop(1,0);
-	} while (tok != ENDINPUT && (tok != LEXERR || isset(SHINSTDIN)));
+	while (tok != ENDINPUT && (tok != LEXERR || isset(SHINSTDIN)));
 	if (tok == LEXERR) {
 	    /* Make sure a parse error exits with non-zero status */
 	    if (!lastval)
