@@ -275,8 +275,7 @@ inputline(void)
 	int flags = ZLRF_HISTORY|ZLRF_NOSETTY;
 	if (isset(IGNOREEOF))
 	    flags |= ZLRF_IGNOREEOF;
-	ingetcline = (char *)zleread(ingetcpmptl, ingetcpmptr, flags,
-				     context);
+	ingetcline = zlereadptr(ingetcpmptl, ingetcpmptr, flags, context);
 	histdone |= HISTFLAG_SETTY;
     }
     if (!ingetcline) {
