@@ -271,7 +271,7 @@ get_usage(void)
 }
 
 
-#if !defined HAVE_WAIT3 || !defined HAVE_GETRUSAGE
+#ifndef HAVE_GETRUSAGE
 /* Update status of process that we have just WAIT'ed for */
 
 /**/
@@ -279,26 +279,17 @@ void
 update_process(Process pn, int status)
 {
     struct timezone dummy_tz;
-#ifdef HAVE_GETRUSAGE
-    struct timeval childs = child_usage.ru_stime;
-    struct timeval childu = child_usage.ru_utime;
-#else
-    long childs = shtms.tms_cstime;
-    long childu = shtms.tms_cutime;
-#endif
+    long childs, childu;
 
+    childs = shtms.tms_cstime;
+    childu = shtms.tms_cutime;
     /* get time-accounting info          */
     get_usage();
     gettimeofday(&pn->endtime, &dummy_tz);  /* record time process exited        */
 
     pn->status = status;                    /* save the status returned by WAIT  */
-#ifdef HAVE_GETRUSAGE
-    dtime(&pn->ti.ru_stime, &childs, &child_usage.ru_stime);
-    dtime(&pn->ti.ru_utime, &childu, &child_usage.ru_utime);
-#else
     pn->ti.st  = shtms.tms_cstime - childs; /* compute process system space time */
     pn->ti.ut  = shtms.tms_cutime - childu; /* compute process user space time   */
-#endif
 }
 #endif
 
@@ -450,7 +441,7 @@ update_job(Job jn)
     if ((isset(NOTIFY) || job == thisjob) && (jn->stat & STAT_LOCKED)) {
 	printjob(jn, !!isset(LONGLISTJOBS), 0);
 	if (zleactive)
-	    zrefresh();
+	    zrefreshptr();
     }
     if (sigtrapped[SIGCHLD] && job != thisjob)
 	dotrap(SIGCHLD);
@@ -565,10 +556,18 @@ printtime(struct timeval *real, child_times_t *ti, char *desc)
 #ifdef HAVE_GETRUSAGE
     double total_time;
 #endif
-    int percent;
+    int percent, desclen;
 
     if (!desc)
+    {
 	desc = "";
+	desclen = 0;
+    }
+    else
+    {
+	desc = dupstring(desc);
+	unmetafy(desc, &desclen);
+    }
 
     /* go ahead and compute these, since almost every TIMEFMT will have them */
     elapsed_time = real->tv_sec + real->tv_usec / 1000000.0;
@@ -715,7 +714,7 @@ printtime(struct timeval *real, child_times_t *ti, char *desc)
 		break;
 #endif
 	    case 'J':
-		fprintf(stderr, "%s", desc);
+		fwrite(desc, sizeof(char), desclen, stderr);
 		break;
 	    case '%':
 		putc('%', stderr);
@@ -865,7 +864,7 @@ printjob(Job jn, int lng, int synch)
 	Process qn;
 
 	if (!synch)
-	    trashzle();
+	    trashzleptr();
 	if (doputnl && !synch)
 	    putc('\n', fout);
 	for (pn = jn->procs; pn;) {
@@ -923,8 +922,14 @@ printjob(Job jn, int lng, int synch)
 			(int)(len - 14 + 2 - strlen(sigmsg(WTERMSIG(pn->status)))), "");
 	    else
 		fprintf(fout, "%-*s", len + 2, sigmsg(WTERMSIG(pn->status)));
-	    for (; pn != qn; pn = pn->next)
-		fprintf(fout, (pn->next) ? "%s | " : "%s", pn->text);
+	    for (; pn != qn; pn = pn->next) {
+		char *txt = dupstring(pn->text);
+		int txtlen;
+		unmetafy(txt, &txtlen);
+		fwrite(txt, sizeof(char), txtlen, fout);
+		if (pn->next)
+		    fputs(" | ", fout);
+	    }
 	    putc('\n', fout);
 	    fline = 0;
 	}
@@ -940,10 +945,10 @@ printjob(Job jn, int lng, int synch)
 
     if ((lng & 4) || (interact && job == thisjob &&
 		      jn->pwd && strcmp(jn->pwd, pwd))) {
-	fprintf(fout, "(pwd %s: ", (lng & 4) ? "" : "now");
-	fprintdir(((lng & 4) && jn->pwd) ? jn->pwd : pwd, fout);
-	fprintf(fout, ")\n");
-	fflush(fout);
+	fprintf(shout, "(pwd %s: ", (lng & 4) ? "" : "now");
+	fprintdir(((lng & 4) && jn->pwd) ? jn->pwd : pwd, shout);
+	fprintf(shout, ")\n");
+	fflush(shout);
     }
 /* delete job if done */
 
@@ -1111,16 +1116,11 @@ havefiles(void)
 
 }
 
-/*
- * Wait for a particular process.
- * wait_cmd indicates this is from the interactive wait command,
- * in which case the behaviour is a little different:  the command
- * itself can be interrupted by a trapped signal.
- */
+/* wait for a particular process */
 
 /**/
-int
-waitforpid(pid_t pid, int wait_cmd)
+void
+waitforpid(pid_t pid)
 {
     int first = 1, q = queue_signal_level();
 
@@ -1133,30 +1133,18 @@ waitforpid(pid_t pid, int wait_cmd)
 	else
 	    kill(pid, SIGCONT);
 
-	last_signal = -1;
-	signal_suspend(SIGCHLD, wait_cmd);
-	if (last_signal != SIGCHLD && wait_cmd) {
-	    /* wait command interrupted, but no error: return */
-	    restore_queue_signals(q);
-	    return 128 + last_signal;
-	}
+	signal_suspend(SIGCHLD, SIGINT);
 	child_block();
     }
     child_unblock();
     restore_queue_signals(q);
-
-    return 0;
 }
 
-/*
- * Wait for a job to finish.
- * wait_cmd indicates this is from the wait builtin; see
- * wait_cmd in waitforpid().
- */
+/* wait for a job to finish */
 
 /**/
-static int
-zwaitjob(int job, int wait_cmd)
+static void
+zwaitjob(int job, int sig)
 {
     int q = queue_signal_level();
     Job jn = jobtab + job;
@@ -1170,13 +1158,7 @@ zwaitjob(int job, int wait_cmd)
 	while (!errflag && jn->stat &&
 	       !(jn->stat & STAT_DONE) &&
 	       !(interact && (jn->stat & STAT_STOPPED))) {
-	    signal_suspend(SIGCHLD, wait_cmd);
-	    if (last_signal != SIGCHLD && wait_cmd)
-	    {
-		/* builtin wait interrupted by trapped signal */
-		restore_queue_signals(q);
-		return 128 + last_signal;
-	    }
+	    signal_suspend(SIGCHLD, sig);
 	    /* Commenting this out makes ^C-ing a job started by a function
 	       stop the whole function again.  But I guess it will stop
 	       something else from working properly, we have to find out
@@ -1199,8 +1181,6 @@ zwaitjob(int job, int wait_cmd)
     }
     child_unblock();
     restore_queue_signals(q);
-
-    return 0;
 }
 
 /* wait for running job to finish */
@@ -1255,15 +1235,6 @@ clearjobtab(int monitor)
 
     memset(jobtab, 0, jobtabsize * sizeof(struct job)); /* zero out table */
     maxjob = 0;
-
-    /*
-     * Although we don't have job control in subshells, we
-     * sometimes needs control structures for other purposes such
-     * as multios.  Grab a job for this purpose; any will do
-     * since we've freed them all up (so there's no question
-     * of problems with the job table size here).
-     */
-    thisjob = initjob();
 }
 
 static int initnewjob(int i)
@@ -1721,9 +1692,9 @@ bin_fg(char *name, char **argv, Options ops, int func)
 	} else {   /* Must be BIN_WAIT, so wait for all jobs */
 	    for (job = 0; job <= maxjob; job++)
 		if (job != thisjob && jobtab[job].stat)
-		    retval = zwaitjob(job, 1);
+		    zwaitjob(job, SIGINT);
 	    unqueue_signals();
-	    return retval;
+	    return 0;
 	}
     }
 
@@ -1741,19 +1712,11 @@ bin_fg(char *name, char **argv, Options ops, int func)
 	    Job j;
 	    Process p;
 
-	    if (findproc(pid, &j, &p, 0)) {
-		/*
-		 * returns 0 for normal exit, else signal+128
-		 * in which case we should return that status.
-		 */
-		retval = waitforpid(pid, 1);
-		if (!retval)
-		    retval = lastval2;
-	    } else {
+	    if (findproc(pid, &j, &p, 0))
+		waitforpid(pid);
+	    else
 		zwarnnam(name, "pid %d is not a child of this shell", 0, pid);
-		/* presumably lastval2 doesn't tell us a heck of a lot? */
-		retval = 1;
-	    }
+	    retval = lastval2;
 	    thisjob = ocj;
 	    continue;
 	}
@@ -1784,21 +1747,9 @@ bin_fg(char *name, char **argv, Options ops, int func)
 	case BIN_WAIT:
 	    if (func == BIN_BG)
 		jobtab[job].stat |= STAT_NOSTTY;
-	    if ((stopped = (jobtab[job].stat & STAT_STOPPED))) {
+	    if ((stopped = (jobtab[job].stat & STAT_STOPPED)))
 		makerunning(jobtab + job);
-		if (func == BIN_BG) {
-		    /* Set $! to indicate this was backgrounded */
-		    Process pn = jobtab[job].procs;
-		    for (;;) {
-			Process next = pn->next;
-			if (!next) {
-			    lastpid = (zlong) pn->pid;
-			    break;
-			}
-			pn = next;
-		    }
-		}
-	    } else if (func == BIN_BG) {
+	    else if (func == BIN_BG) {
 		/* Silly to bg a job already running. */
 		zwarnnam(name, "job already in background", NULL, 0);
 		thisjob = ocj;
@@ -1822,12 +1773,11 @@ bin_fg(char *name, char **argv, Options ops, int func)
 		printjob(jobtab + job, (stopped) ? -1 : lng, 1);
 	    if (func != BIN_BG) {		/* fg or wait */
 		if (jobtab[job].pwd && strcmp(jobtab[job].pwd, pwd)) {
-		    FILE *fout = (func == BIN_JOBS) ? stdout : shout;
-		    fprintf(fout, "(pwd : ");
-		    fprintdir(jobtab[job].pwd, fout);
-		    fprintf(fout, ")\n");
-		    fflush(fout);
+		    fprintf(shout, "(pwd : ");
+		    fprintdir(jobtab[job].pwd, shout);
+		    fprintf(shout, ")\n");
 		}
+		fflush(shout);
 		if (func != BIN_WAIT) {		/* fg */
 		    thisjob = job;
 		    if ((jobtab[job].stat & STAT_SUPERJOB) &&
@@ -1846,18 +1796,8 @@ bin_fg(char *name, char **argv, Options ops, int func)
 		killjb(jobtab + job, SIGCONT);
 	    }
 	    if (func == BIN_WAIT)
-	    {
-		retval = zwaitjob(job, 1);
-		if (!retval)
-		    retval = lastval2;
-	    }
-	    else if (func != BIN_BG) {
-		/*
-		 * HERE: there used not to be an "else" above.  How
-		 * could it be right to wait for the foreground job
-		 * when we've just been told to wait for another
-		 * job (and done it)?
-		 */
+		zwaitjob(job, SIGINT);
+	    if (func != BIN_BG) {
 		waitjobs();
 		retval = lastval2;
 	    } else if (ofunc == BIN_DISOWN)
@@ -1906,18 +1846,6 @@ bin_fg(char *name, char **argv, Options ops, int func)
     return retval;
 }
 
-#if defined(SIGCHLD) && defined(SIGCLD)
-#if SIGCHLD == SIGCLD
-#define ALT_SIGS 1
-#endif
-#endif
-#if defined(SIGPOLL) && defined(SIGIO)
-#if SIGPOLL == SIGIO
-#define ALT_SIGS 1
-#endif
-#endif
-
-#ifdef ALT_SIGS
 const struct {
     const char *name;
     int num;
@@ -1931,10 +1859,16 @@ const struct {
 #if SIGPOLL == SIGIO
     { "IO", SIGIO },
 #endif
+#endif
+#if !defined(SIGERR)
+    /*
+     * If SIGERR is not defined by the operating system, use it
+     * as an alias for SIGZERR.
+     */
+    { "ERR", SIGZERR },
 #endif
     { NULL, 0 }
 };
-#endif
 
 /* kill: send a signal to a process.  The process(es) may be specified *
  * by job specifier (see above) or pid.  A signal, defaulting to       *
@@ -1966,7 +1900,6 @@ bin_kill(char *nam, char **argv, UNUSED(Options ops), UNUSED(int func))
 			    for (sig = 1; sig <= SIGCOUNT; sig++)
 				if (!cstrpcmp(sigs + sig, &signame))
 				    break;
-#ifdef ALT_SIGS
 			    if (sig > SIGCOUNT) {
 				int i;
 
@@ -1977,7 +1910,6 @@ bin_kill(char *nam, char **argv, UNUSED(Options ops), UNUSED(int func))
 					break;
 				    }
 			    }
-#endif
 			    if (sig > SIGCOUNT) {
 				zwarnnam(nam, "unknown signal: SIG%s",
 					 signame, 0);
@@ -2040,7 +1972,6 @@ bin_kill(char *nam, char **argv, UNUSED(Options ops), UNUSED(int func))
 			break;
 		if (*signame == '0' && !signame[1])
 		    sig = 0;
-#ifdef ALT_SIGS
 		if (sig > SIGCOUNT) {
 		    int i;
 
@@ -2051,7 +1982,6 @@ bin_kill(char *nam, char **argv, UNUSED(Options ops), UNUSED(int func))
 			    break;
 			}
 		}
-#endif
 		if (sig > SIGCOUNT) {
 		    zwarnnam(nam, "unknown signal: SIG%s", signame, 0);
 		    zwarnnam(nam, "type kill -l for a List of signals", NULL, 0);
@@ -2129,18 +2059,40 @@ getsignum(char *s)
 	if (!strcmp(s, sigs[i]))
 	    return i;
 
-#ifdef ALT_SIGS
     for (i = 0; alt_sigs[i].name; i++)
     {
 	if (!strcmp(s, alt_sigs[i].name))
 	    return alt_sigs[i].num;
     }
-#endif
 
     /* no matching signal */
     return -1;
 }
 
+/* Get the name for a signal. */
+
+/**/
+mod_export const char *
+getsigname(int sig)
+{
+    if (sigtrapped[sig] & ZSIG_ALIAS)
+    {
+	int i;
+	for (i = 0; alt_sigs[i].name; i++)
+	    if (sig == alt_sigs[i].num)
+		return alt_sigs[i].name;
+    }
+    else
+	return sigs[sig];
+
+    /* shouldn't reach here */
+#ifdef DEBUG
+    dputs("Bad alias flag for signal");
+#endif
+    return "";
+}
+
+
 /* Get the function node for a trap, taking care about alternative names */
 /**/
 HashNode
@@ -2149,9 +2101,7 @@ gettrapnode(int sig, int ignoredisable)
     char fname[20];
     HashNode hn;
     HashNode (*getptr)(HashTable ht, char *name);
-#ifdef ALT_SIGS
     int i;
-#endif
     if (ignoredisable)
 	getptr = shfunctab->getnode2;
     else
@@ -2161,7 +2111,6 @@ gettrapnode(int sig, int ignoredisable)
     if ((hn = getptr(shfunctab, fname)))
 	return hn;
 
-#ifdef ALT_SIGS
     for (i = 0; alt_sigs[i].name; i++) {
 	if (alt_sigs[i].num == sig) {
 	    sprintf(fname, "TRAP%s", alt_sigs[i].name);
@@ -2169,7 +2118,6 @@ gettrapnode(int sig, int ignoredisable)
 		return hn;
 	}
     }
-#endif
 
     return NULL;
 }
