@@ -33,7 +33,7 @@
 /* current emulation (used to decide which set of option letters is used) */
 
 /**/
-int emulation;
+mod_export int emulation;
  
 /* the options; e.g. if opts[SHGLOB] != 0, SH_GLOB is turned on */
  
@@ -138,6 +138,7 @@ static struct optname optns[] = {
 {NULL, "histnofunctions",     0,			 HISTNOFUNCTIONS},
 {NULL, "histnostore",	      0,			 HISTNOSTORE},
 {NULL, "histreduceblanks",    0,			 HISTREDUCEBLANKS},
+{NULL, "histsavebycopy",      OPT_ALL,			 HISTSAVEBYCOPY},
 {NULL, "histsavenodups",      0,			 HISTSAVENODUPS},
 {NULL, "histverify",	      0,			 HISTVERIFY},
 {NULL, "hup",		      OPT_EMULATE|OPT_ZSH,	 HUP},
@@ -180,6 +181,7 @@ static struct optname optns[] = {
 {NULL, "promptbang",	      OPT_KSH,			 PROMPTBANG},
 {NULL, "promptcr",	      OPT_ALL,			 PROMPTCR},
 {NULL, "promptpercent",	      OPT_NONBOURNE,		 PROMPTPERCENT},
+{NULL, "promptsp",	      OPT_ALL,			 PROMPTSP},
 {NULL, "promptsubst",	      OPT_KSH,			 PROMPTSUBST},
 {NULL, "pushdignoredups",     OPT_EMULATE,		 PUSHDIGNOREDUPS},
 {NULL, "pushdminus",	      OPT_EMULATE,		 PUSHDMINUS},
@@ -209,6 +211,7 @@ static struct optname optns[] = {
 {NULL, "unset",		      OPT_EMULATE|OPT_BSHELL,	 UNSET},
 {NULL, "verbose",	      0,			 VERBOSE},
 {NULL, "vi",		      0,			 VIMODE},
+{NULL, "warncreateglobal",    0,			 WARNCREATEGLOBAL},
 {NULL, "xtrace",	      0,			 XTRACE},
 {NULL, "zle",		      OPT_SPECIAL,		 USEZLE},
 {NULL, "braceexpand",	      OPT_ALIAS, /* ksh/bash */	 -IGNOREBRACES},
@@ -589,14 +592,7 @@ optlookup(char const *name)
 	if (*t == '_')
 	    chuck(t);
 	else {
-	    /*
-	     * Some locales (in particular tr_TR.UTF-8) may
-	     * have non-standard mappings of ASCII characters,
-	     * so be careful.  Option names must be ASCII so
-	     * we don't need to be too clever.
-	     */
-	    if (*t >= 'A' && *t <= 'Z')
-		*t = (*t - 'A') + 'a';
+	    *t = tulower(*t);
 	    t++;
 	}
 
@@ -683,16 +679,7 @@ dosetopt(int optno, int value, int force)
 	setuid(getuid());
 	setgid(getgid());
 #endif /* HAVE_SETUID */
-#ifdef JOB_CONTROL
-    } else if (!force && optno == MONITOR && value) {
-	if (opts[optno] == value)
-	    return 0;
-	if (interact && (SHTTY != -1)) {
-	    origpgrp = GETPGRP();
-	    acquire_pgrp();
-	} else
-	    return -1;
-#else
+#ifndef JOB_CONTROL
     } else if(optno == MONITOR && value) {
 	    return -1;
 #endif /* not JOB_CONTROL */
