@@ -33,14 +33,6 @@
 static char *tptr, *tbuf, *tlim;
 static int tsiz, tindent, tnewlins, tjob;
 
-static void
-dec_tindent(void)
-{
-    DPUTS(tindent == 0, "attempting to decrement tindent below zero");
-    if (tindent > 0)
-	tindent--;
-}
-
 /* add a character to the text buffer */
 
 /**/
@@ -362,7 +354,7 @@ gettext2(Estate state)
 		state->pc++;
 	    } else {
 		state->pc = s->u._subsh.end;
-		dec_tindent();
+		tindent--;
 		taddnl();
 		taddstr(")");
 		stack = 1;
@@ -379,7 +371,7 @@ gettext2(Estate state)
 		state->pc++;
 	    } else {
 		state->pc = s->u._subsh.end;
-		dec_tindent();
+		tindent--;
 		taddnl();
 		taddstr("}");
 		stack = 1;
@@ -395,7 +387,7 @@ gettext2(Estate state)
 		} else
 		    stack = 1;
 	    } else {
-		dec_tindent();
+		tindent--;
 		stack = 1;
 	    }
 	    break;
@@ -422,7 +414,7 @@ gettext2(Estate state)
 	    } else {
 		state->strs = s->u._funcdef.strs;
 		state->pc = s->u._funcdef.end;
-		dec_tindent();
+		tindent--;
 		taddnl();
 		taddstr("}");
 		stack = 1;
@@ -452,7 +444,7 @@ gettext2(Estate state)
 		taddnl();
 		tpush(code, 1);
 	    } else {
-		dec_tindent();
+		tindent--;
 		taddnl();
 		taddstr("done");
 		stack = 1;
@@ -470,7 +462,7 @@ gettext2(Estate state)
 		taddnl();
 		tpush(code, 1);
 	    } else {
-		dec_tindent();
+		tindent--;
 		taddnl();
 		taddstr("done");
 		stack = 1;
@@ -483,14 +475,14 @@ gettext2(Estate state)
 		tindent++;
 		tpush(code, 0);
 	    } else if (!s->pop) {
-		dec_tindent();
+		tindent--;
 		taddnl();
 		taddstr("do");
 		tindent++;
 		taddnl();
 		s->pop = 1;
 	    } else {
-		dec_tindent();
+		tindent--;
 		taddnl();
 		taddstr("done");
 		stack = 1;
@@ -506,7 +498,7 @@ gettext2(Estate state)
 		taddnl();
 		tpush(code, 1);
 	    } else {
-		dec_tindent();
+		tindent--;
 		taddnl();
 		taddstr("done");
 		stack = 1;
@@ -544,20 +536,8 @@ gettext2(Estate state)
 		    n->pop = (state->pc - 2 + WC_CASE_SKIP(code) >= end);
 		}
 	    } else if (state->pc < s->u._case.end) {
-		dec_tindent();
-		switch (WC_CASE_TYPE(code)) {
-		case WC_CASE_OR:
-		    taddstr(" ;;");
-		    break;
-
-		case WC_CASE_AND:
-		    taddstr(";&");
-		    break;
-
-		default:
-		    taddstr(";|");
-		    break;
-		}
+		tindent--;
+		taddstr(WC_CASE_TYPE(code) == WC_CASE_OR ? " ;;" : ";&");
 		if (tnewlins)
 		    taddnl();
 		else
@@ -572,21 +552,9 @@ gettext2(Estate state)
 		s->pop = ((state->pc - 2 + WC_CASE_SKIP(code)) >=
 			  s->u._case.end);
 	    } else {
-		dec_tindent();
-		switch (WC_CASE_TYPE(code)) {
-		case WC_CASE_OR:
-		    taddstr(" ;;");
-		    break;
-
-		case WC_CASE_AND:
-		    taddstr(";&");
-		    break;
-
-		default:
-		    taddstr(";|");
-		    break;
-		}
-		dec_tindent();
+		tindent--;
+		taddstr(WC_CASE_TYPE(code) == WC_CASE_OR ? " ;;" : ";&");
+		tindent--;
 		if (tnewlins)
 		    taddnl();
 		else
@@ -609,14 +577,14 @@ gettext2(Estate state)
 	    } else if (s->pop) {
 		stack = 1;
 	    } else if (s->u._if.cond) {
-		dec_tindent();
+		tindent--;
 		taddnl();
 		taddstr("then");
 		tindent++;
 		taddnl();
 		s->u._if.cond = 0;
 	    } else if (state->pc < s->u._if.end) {
-		dec_tindent();
+		tindent--;
 		taddnl();
 		code = *state->pc++;
 		if (WC_IF_TYPE(code) == WC_IF_ELIF) {
@@ -630,7 +598,7 @@ gettext2(Estate state)
 		}
 	    } else {
 		s->pop = 1;
-		dec_tindent();
+		tindent--;
 		taddnl();
 		taddstr("fi");
 		stack = 1;
@@ -768,14 +736,14 @@ gettext2(Estate state)
 		n->u._subsh.end = state->pc + WC_CURSH_SKIP(state->pc[-1]);
 	    } else if (!s->pop) {
 		state->pc = s->u._subsh.end;
-		dec_tindent();
+		tindent--;
 		taddnl();
 		taddstr("} always {");
 		tindent++;
 		taddnl();
 		s->pop = 1;
 	    } else {
-		dec_tindent();
+		tindent--;
 		taddnl();
 		taddstr("}");
 		stack = 1;
@@ -821,21 +789,26 @@ getredirs(LinkList redirs)
 	case REDIR_MERGEOUT:
 	case REDIR_INPIPE:
 	case REDIR_OUTPIPE:
-	    if (f->fd1 != (IS_READFD(f->type) ? 0 : 1))
+	    if (f->varid) {
+		taddchr('{');
+		taddstr(f->varid);
+		taddchr('}');
+	    } else if (f->fd1 != (IS_READFD(f->type) ? 0 : 1))
 		taddchr('0' + f->fd1);
 	    taddstr(fstr[f->type]);
-	    taddchr(' ');
- 	    if (f->type == REDIR_HERESTR && !has_token(f->name)) {
- 		/*
- 		 * Strings that came from here-documents are converted
- 		 * to here strings without quotation, so add that
- 		 * now.  If tokens are already present taddstr()
- 		 * will do the right thing (anyway, adding more
- 		 * quotes certainly isn't right in that case).
- 		 */
- 		taddchr('\'');
- 		taddstr(bslashquote(f->name, NULL, 1));
- 		taddchr('\'');
+	    if (f->type != REDIR_MERGEIN && f->type != REDIR_MERGEOUT)
+		taddchr(' ');
+	    if (f->type == REDIR_HERESTR && !has_token(f->name)) {
+		/*
+		 * Strings that came from here-documents are converted
+		 * to here strings without quotation, so add that
+		 * now.  If tokens are already present taddstr()
+		 * will do the right thing (anyway, adding more
+		 * quotes certainly isn't right in that case).
+		 */
+		taddchr('\'');
+		taddstr(bslashquote(f->name, NULL, 1));
+		taddchr('\'');
 	    } else
 		taddstr(f->name);
 	    taddchr(' ');
