@@ -27,14 +27,6 @@
  *
  */
 
-#define trashzle()      trashzleptr()
-#define zle_resetprompt()      zle_resetpromptptr()
-#define zleread(X,Y,H,C)  zlereadptr(X,Y,H,C)
-#define spaceinline(X)  spaceinlineptr(X)
-#define zrefresh()      refreshptr()
-
-#define compctlread(N,A,O,R) compctlreadptr(N,A,O,R)
-
 /* A few typical macros */
 #define minimum(a,b)  ((a) < (b) ? (a) : (b))
 
@@ -171,8 +163,6 @@ struct mathfunc {
  */
 #define Nularg		((char) 0x9c)
 
-#define INULL(x)	((x) >= Snull && (x) <= Nularg)
-
 /*
  * Take care to update the use of IMETA appropriately when adding
  * tokens here.
@@ -217,38 +207,37 @@ enum {
     DOUTPAR,
     AMPERBANG,		/* 30 */
     SEMIAMP,
-    SEMIBAR,
     DOUTBRACK,
     STRING,
-    ENVSTRING,		/* 35 */
-    ENVARRAY,
+    ENVSTRING,
+    ENVARRAY,		/* 35 */
     ENDINPUT,
     LEXERR,
 
     /* Tokens for reserved words */
     BANG,	/* !         */
-    DINBRACK,	/* [[        */	/* 40 */
-    INBRACE,    /* {         */
+    DINBRACK,	/* [[        */
+    INBRACE,    /* {         */	/* 40 */
     OUTBRACE,   /* }         */
     CASE,	/* case      */
     COPROC,	/* coproc    */
-    DOLOOP,	/* do        */ /* 45 */
-    DONE,	/* done      */
+    DOLOOP,	/* do        */
+    DONE,	/* done      */ /* 45 */
     ELIF,	/* elif      */
     ELSE,	/* else      */
     ZEND,	/* end       */
-    ESAC,	/* esac      */ /* 50 */
-    FI,		/* fi        */
+    ESAC,	/* esac      */
+    FI,		/* fi        */ /* 50 */
     FOR,	/* for       */
     FOREACH,	/* foreach   */
     FUNC,	/* function  */
-    IF,		/* if        */ /* 55 */
-    NOCORRECT,	/* nocorrect */
+    IF,		/* if        */
+    NOCORRECT,	/* nocorrect */ /* 55 */
     REPEAT,	/* repeat    */
     SELECT,	/* select    */
     THEN,	/* then      */
-    TIME,	/* time      */ /* 60 */
-    UNTIL,	/* until     */
+    TIME,	/* time      */
+    UNTIL,	/* until     */ /* 60 */
     WHILE	/* while     */
 };
 
@@ -276,6 +265,8 @@ enum {
     REDIR_INPIPE,		/* < <(...) */
     REDIR_OUTPIPE		/* > >(...) */
 };
+#define REDIR_TYPE_MASK	(0x1f)
+#define REDIR_VARID_MASK (0x20)
 
 #define IS_WRITE_FILE(X)      ((X)>=REDIR_WRITE && (X)<=REDIR_READWRITE)
 #define IS_APPEND_REDIR(X)    (IS_WRITE_FILE(X) && ((X) & 2))
@@ -296,10 +287,15 @@ enum {
  * processes.
  */
 #define FDT_INTERNAL		1
+/*
+ * Entry visible to other processes, for example created using
+ * the {varid}> file syntax.
+ */
+#define FDT_EXTERNAL		2
 /*
  * Entry used by output from the XTRACE option.
  */
-#define FDT_XTRACE		2
+#define FDT_XTRACE		3
 #ifdef PATH_DEV_FD
 /*
  * Entry used by a process substition.
@@ -307,7 +303,7 @@ enum {
  * decremented on exit; we don't close entries greater than
  * FDT_PROC_SUBST except when closing everything.
  */
-#define FDT_PROC_SUBST		3
+#define FDT_PROC_SUBST		4
 #endif
 
 /* Flags for input stack */
@@ -483,6 +479,7 @@ struct redir {
     int type;
     int fd1, fd2;
     char *name;
+    char *varid;
 };
 
 /* The number of fds space is allocated for  *
@@ -659,8 +656,11 @@ struct eccstr {
 #define WC_PIPE_LINENO(C)   (wc_data(C) >> 1)
 #define WCB_PIPE(T,L)       wc_bld(WC_PIPE, ((T) | ((L) << 1)))
 
-#define WC_REDIR_TYPE(C)    wc_data(C)
+#define WC_REDIR_TYPE(C)    ((int)(wc_data(C) & REDIR_TYPE_MASK))
+#define WC_REDIR_VARID(C)   ((int)(wc_data(C) & REDIR_VARID_MASK))
 #define WCB_REDIR(T)        wc_bld(WC_REDIR, (T))
+/* Size of redir is 4 words if REDIR_VARID_MASK is set, else 3 */
+#define WC_REDIR_WORDS(C)   (WC_REDIR_VARID(C) ? 4 : 3)
 
 #define WC_ASSIGN_TYPE(C)   (wc_data(C) & ((wordcode) 1))
 #define WC_ASSIGN_TYPE2(C)  ((wc_data(C) & ((wordcode) 2)) >> 1)
@@ -713,14 +713,12 @@ struct eccstr {
 #define WC_TRY_SKIP(C)	    wc_data(C)
 #define WCB_TRY(O)	    wc_bld(WC_TRY, (O))
 
-#define WC_CASE_TYPE(C)     (wc_data(C) & 7)
+#define WC_CASE_TYPE(C)     (wc_data(C) & 3)
 #define WC_CASE_HEAD        0
 #define WC_CASE_OR          1
 #define WC_CASE_AND         2
-#define WC_CASE_TESTAND     3
-#define WC_CASE_FREE	    (3) /* Next bit available in integer */
-#define WC_CASE_SKIP(C)     (wc_data(C) >> WC_CASE_FREE)
-#define WCB_CASE(T,O)       wc_bld(WC_CASE, ((T) | ((O) << WC_CASE_FREE)))
+#define WC_CASE_SKIP(C)     (wc_data(C) >> 2)
+#define WCB_CASE(T,O)       wc_bld(WC_CASE, ((T) | ((O) << 2)))
 
 #define WC_IF_TYPE(C)       (wc_data(C) & 3)
 #define WC_IF_HEAD          0
@@ -982,8 +980,6 @@ struct shfunc {
 struct funcstack {
     Funcstack prev;		/* previous in stack */
     char *name;			/* name of function called */
-    char *caller;		/* name of caller */
-    int lineno;			/* line number in file */
 };
 
 /* node in list of function call wrappers */
@@ -1384,6 +1380,14 @@ struct paramdef {
 #define setsparam(S,V) assignsparam(S,V,0)
 #define setaparam(S,V) assignaparam(S,V,0)
 
+/*
+ * Flags for assignsparam and assignaparam.
+ */
+enum {
+    ASSPM_AUGMENT = 1 << 0,
+    ASSPM_WARN_CREATE = 1 << 1
+};
+
 /* node for named directory hash table (nameddirtab) */
 
 struct nameddir {
@@ -1428,7 +1432,12 @@ struct histent {
 
     Histent up;			/* previous line (moving upward)    */
     Histent down;		/* next line (moving downward)      */
+#ifdef MULTIBYTE_SUPPORT	/* (Note: must match ZLE_STRING_T!) */
+    wchar_t *zle_text;		/* the edited history line          */
+#else
     char *zle_text;		/* the edited history line          */
+#endif
+    int zle_len;		/* length of zle_text */
     time_t stim;		/* command started time (datestamp) */
     time_t ftim;		/* command finished time            */
     short *words;		/* Position of words in history     */
@@ -1569,6 +1578,7 @@ enum {
     HISTNOFUNCTIONS,
     HISTNOSTORE,
     HISTREDUCEBLANKS,
+    HISTSAVEBYCOPY,
     HISTSAVENODUPS,
     HISTVERIFY,
     HUP,
@@ -1611,6 +1621,7 @@ enum {
     PROMPTBANG,
     PROMPTCR,
     PROMPTPERCENT,
+    PROMPTSP,
     PROMPTSUBST,
     PUSHDIGNOREDUPS,
     PUSHDMINUS,
@@ -1640,6 +1651,7 @@ enum {
     UNSET,
     VERBOSE,
     VIMODE,
+    WARNCREATEGLOBAL,
     XTRACE,
     USEZLE,
     DVORAK,
@@ -1690,9 +1702,7 @@ struct ttyinfo {
 #  ifdef OXTABS
 #define SGTABTYPE       OXTABS
 #  else
-#   ifdef XTABS
 #define SGTABTYPE       XTABS
-#   endif
 #  endif
 # endif
 
@@ -1735,7 +1745,10 @@ struct ttyinfo {
 #define TCDOWNCURSOR   26
 #define TCLEFTCURSOR   27
 #define TCRIGHTCURSOR  28
-#define TC_COUNT       29
+#define TCSAVECURSOR   29
+#define TCRESTRCURSOR  30
+#define TCBACKSPACE    31
+#define TC_COUNT       32
 
 #define tccan(X) (tclen[X])
 
@@ -1871,7 +1884,8 @@ struct heap {
 /* Mask to get the above flags */
 #define ZSIG_MASK	(ZSIG_TRAPPED|ZSIG_IGNORED|ZSIG_FUNC)
 /* No. of bits to shift local level when storing in sigtrapped */
-#define ZSIG_SHIFT	3
+#define ZSIG_ALIAS	(1<<3)  /* Trap is stored under an alias */
+#define ZSIG_SHIFT	4
 
 /**********************************/
 /* Flags to third argument of zle */
@@ -1904,7 +1918,8 @@ typedef int (*CompctlReadFn) _((char *, char **, Options, char *));
 
 typedef void (*ZleVoidFn) _((void));
 typedef void (*ZleVoidIntFn) _((int));
-typedef unsigned char * (*ZleReadFn) _((char **, char **, int, int));
+typedef char *(*ZleReadFn) _((char **, char **, int, int));
+typedef char *(*ZleGetLineFn) _((int *, int *));
 
 /***************************************/
 /* Hooks in core.                      */
@@ -1913,3 +1928,10 @@ typedef unsigned char * (*ZleReadFn) _((char **, char **, int, int));
 #define EXITHOOK       (zshhooks + 0)
 #define BEFORETRAPHOOK (zshhooks + 1)
 #define AFTERTRAPHOOK  (zshhooks + 2)
+
+#ifdef MULTIBYTE_SUPPORT
+#define nicezputs(str, outs)	(void)mb_niceformat((str), (outs), NULL, 0)
+
+#define MB_INCOMPLETE	((size_t)-2)
+#define MB_INVALID	((size_t)-1)
+#endif
